 #!/bin/bash
set -v
set -x
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# This script tries to simplify migration from 0.92 to 0.921.
# Because of the folder restructuration

create_sgd_folder () {
if [ ! -d "$1" ] ; then
mkdir $1
cp ${MY_BLANK_FILE} ./${1}/sdg_module
fi
}

echo "Be careful. This script must be run from dev folder"
echo 
SDG_FOLDERS_TO_MIGRATE="common lng S10en S20es S21ca S25eu S30deu S30fra S30hun";
MY_BLANK_FILE=`pwd`"/util/blank_file"

for n_migrated_folder in $SDG_FOLDERS_TO_MIGRATE; do
if [ -d "$n_migrated_folder" ] ;   then #We check that the folder exists
cd $n_migrated_folder 

# 1st of all create new folders
create_sgd_folder S10_linux
create_sgd_folder S60_advanced
# inside advanced
cd S60_advanced
	create_sgd_folder S10_grub
	create_sgd_folder S15_lilo
	create_sgd_folder S20_a_linux
	create_sgd_folder S30_a_windows
	create_sgd_folder S40_other_oses
	#create_sgd_folder S50_special_boot
	#create_sgd_folder S60_setup (To be moved from root)
	create_sgd_folder S70_misc
cd .. # Go to root

# GNU/LINUX BASIC are links ... Create it manually.
# Windows is already ok.... Manually edit it in a better place.
# Backup... NOT TO BE CREATED YET
# Tools... Here we have to copy the classic boot options here and delete the classic boot.
#
if [ -d S40_usualboot ] ; then
  cp -a S40_usualboot/* S50_utils/
  rm -R S40_usualboot
fi

# Grub menu
# We need the Restore Grub to the MBR option
if [ -d S10_insmbr ] ; then
  mv S10_insmbr S60_advanced/S10_grub/
fi
# We need the option for loading configfile from the hard disk
if [ -d S30_specialboot/S10boot_youros ] ; then
  mv S30_specialboot/S10boot_youros S60_advanced/S10_grub/
fi

# Uninstall Grub Link to be created manually
# Gnu/Linux Advanced: Links to be created manually
# Gnu/Linux Advanced: The option for booting Gnu/Linux manually

if [ -d S25_bootoses/S25_linux ] ; then
  mv S25_bootoses/S25_linux S60_advanced/S20_a_linux/
fi


# Windows Advanced: Links to be created manually
# But we have an option: Fix Boot of Windows (partition)
# 
# It is not an option we will have to write it directly on the body_menu.lst
# No need to move anything

# Other oses
if [ -d S25_bootoses/S50_hurd ] ; then
  mv S25_bootoses/S50_hurd S60_advanced/S40_other_oses/
fi

if [ -d S25_bootoses/S75_osolaris ] ; then
  mv S25_bootoses/S75_osolaris S60_advanced/S40_other_oses/
fi

if [ -d S30_specialboot ] ; then
  mv S30_specialboot S60_advanced/S50_special_boot
fi

if [ -d S60_setup ] ; then
  mv S60_setup S60_advanced/
fi

# I think this is all. I'll have some problems with Linux folder.

# Some things I had forgotten
if [ -d S15_inslilo ] ; then
  mv S15_inslilo S60_advanced/S15_lilo/
fi

if [ -d S25_bootoses ] ; then
  rm -R S25_bootoses
fi

if [ -d S75_uninsgrub ] ; then
  mv S75_uninsgrub S20_bootwin/
fi

if [ -d S60_advanced/S20_a_linux/S25_linux ] ; then
  mv S60_advanced/S20_a_linux/S25_linux S60_advanced/S70_misc
fi

if [ -d S60_advanced/S20_a_linux/S25_linux ] ; then
  mv S60_advanced/S20_a_linux/S25_linux S60_advanced/S70_misc
fi

if [ -d S20_bootwin/S75_uninsgrub ] ; then
  mv S20_bootwin/S75_uninsgrub S60_advanced/S70_misc
fi

if [ -d S50_utils/S50_fixwpart ] ; then
  mv S50_utils/S50_fixwpart S60_advanced/S70_misc
fi

if [ -d S50_utils/hide ] ; then
  mv S50_utils/hide S50_utils/T20_hide
fi

if [ -d S50_utils/unhide ] ; then
  mv S50_utils/unhide S50_utils/T30_unhide
fi

if [ -d S50_utils/S10_actipart ] ; then
  mv S50_utils/S10_actipart S50_utils/T10_actipart
fi


# We do not need to rename the boot options
if [ -d S60_advanced/S70_misc/S75_uninsgrub ] ; then
  mv S60_advanced/S70_misc/S75_uninsgrub S60_advanced/S10_grub/
fi



cd .. # Next n_migrated_folder
fi
done;