#!/bin/bash
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

set -x
set -v


# options: 
# Build a cdrom iso


clean_target()
{
# Prepare target folder
MAGIC_TARGET="./test_fs"


if [ -d  $MAGIC_TARGET ] ; then
  # If target folder exists we should delete it and create it again.
  rm -Rf $MAGIC_TARGET
  mkdir -p $MAGIC_TARGET
else
  # If target folder does exit we create it.
  mkdir -p $MAGIC_TARGET;
fi
}

usage()
{
        cat <<EOF
Usage: $0 [options] cdromiso /path/to/iso
Creates ${MAGIC_NAME} cdrom iso in /path/to/iso file.

Usage: $0 [options] floppy
Creates ${MAGIC_NAME} floppy image in /path/to/floppy file.

Usage: $0 [options] fs
Creates a ${MAGIC_NAME} filesystem in test_fs without copying /boot.

options:
        -h, --help      print this message and exit

EOF
        exit 1
}


cdrom_setup()
{
# Copy /boot contents to to FS_TARGET
if cp ${STAGE2_ELTORITO_FILE_PATH} ${FS_TARGET}/boot/grub/stage2_eltorito ; then
echo -e "GREAT!: We could copy ${STAGE2_ELTORITO_FILE_PATH} file into ${FS_TARGET} folder\n";
else
echo -e "ERROR: We could NOT copy ${STAGE2_ELTORITO_FILE_PATH} file into ${FS_TARGET} folder\n";
exit 1;
fi
}

floppy_setup()
{
# Copy /boot contents to to FS_TARGET
if cp ${STAGE2_FILE_PATH} ${FS_TARGET}/boot/grub/ ; then
echo -e "GREAT!: We could copy ${STAGE2_FILE_PATH} file into ${FS_TARGET} folder\n";
else
echo -e "ERROR: We could NOT copy ${STAGE2_FILE_PATH} file into ${FS_TARGET} folder\n";
exit 1;
fi

if cp ${STAGE1_FILE_PATH} ${FS_TARGET}/boot/grub/ ; then
echo -e "GREAT!: We could copy ${STAGE1_FILE_PATH} file into ${FS_TARGET} folder\n";
else
echo -e "ERROR: We could NOT copy ${STAGE1_FILE_PATH} file into ${FS_TARGET} folder\n";
exit 1;
fi

}


common_setup()
{
mkdir -p ${FS_TARGET}/boot/grub
# Update root menu.lst
cp ${MAGIC_FOLDER}/* ${MAGIC_TARGET}/boot/grub/

}


##############################################################
#########           MAIN PROGRAM	######################
##############################################################

MAGIC_FOLDER="dev_asgd"
MAGIC_TARGET="test_fs"
FS_TARGET="${MAGIC_TARGET}"
MAGIC_NAME="Super Grub Disk"
MAGIC_RELEASE="release"
STAGE2_ELTORITO_FILE_PATH="./boot/grub/stage2_eltorito"
STAGE1_FILE_PATH="./boot/grub/stage1"
STAGE2_FILE_PATH="./boot/grub/stage2"
#STAGE15_FILE_PATH="./boot/grub/fat_stage1_5"

# Check magic folder

	if [ ! -d "$MAGIC_FOLDER" ] ;   then
	echo -e "ERROR: $MAGIC_FOLDER folder is inaccesible\n";
	echo -e "ABORTING!";
	exit 1;
	fi

# Get data from configure
#cd "${MAGIC_FOLDER}"
#source ./sgrub_configure.sh


case "$1" in
        -h | --help)
                usage
                ;;
	cdromiso)
		if [ ! -n "$2" ] ; then
		echo -e "ERROR:\n";
		echo -e "Please specify the ISO where you want to store SGD\n"
		echo -e "e.g. \n"
		echo -e "$0 cdromiso /tmp/sgd_0.xx.iso\n"
		exit 1
		else
		clean_target
		common_setup
		cdrom_setup
			if mkisofs -R -b boot/grub/stage2_eltorito -no-emul-boot -boot-load-size 4 -boot-info-table -o "$2" "$FS_TARGET" ;
			then
			echo -e "GREAT!: We could make the iso file with mkisofs from $FS_TARGET folder\n";
			else
			echo -e "ERROR: We could NOT make the iso file with mkisofs from $FS_TARGET folder. Check permissions.\n";
			exit 1;
			fi

		fi
		;;

	floppy)
		if [ ! -n "$2" ] ; then
		echo -e "ERROR:\n";
		echo -e "Please specify the IMG where you want to store SGD\n"
		echo -e "e.g. \n"
		echo -e "$0 floppy /tmp/sgd_0.xx.img\n"
		exit 1
		else
		clean_target
		common_setup
		floppy_setup
			TEMPORAL_TAR=`pwd`"/./tmp/temp.tar"
			MYCURRENTDIR=`pwd`
			cd "$FS_TARGET" # tar its a little stupid. So we go the correct folder.
			tar cvf "$TEMPORAL_TAR" *
			cd "$MYCURRENTDIR" # Come back to former folder.
			#FIX ME - THIS HAS TO BE INDEPENDENT OF DEV_SGD AND IT HAS TO BE COMMON TO: dev_sgd and dev_asgd.
			#FIX ME- TAKING ABOUT mkfloppyimage.sh script
			sudo ./dev_sgd/mkfloppyimage.sh -f "$TEMPORAL_TAR" -d ./tmp -t 1.44 -s vfat -i "$2"
			sudo chown `id -un`:`id -gn` "$2"

		fi
		;;


#	fs)
#		./build_cdrom_fs.sh
#		;;



        *)
		usage
                ;;
esac
