#!/bin/bash
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#set -x
#set -v


MAGIC_DISTRONAME="SGDPlusKNX"
MAGIC_NAME="${MAGIC_DISTRONAME}"
FS_TARGET="./test_fs"

usage()
{
        cat <<EOF
Usage: $0 [options] cdromiso /path/to/iso
Creates ${MAGIC_NAME} cdrom iso in /path/to/iso file.

Usage: $0 [options] fakecdromiso /path/to/iso
Creates a fake (only menues no root contents) ${MAGIC_NAME} cdrom iso in /path/to/iso file.

Usage: $0 [options] qemutest
Creates ${MAGIC_NAME} cdrom temporarily and runs it in qemu

Usage: $0 [options] fakeqemutest
Creates a fake (only menues no root contents) ${MAGIC_NAME} cdrom temporarily and runs it in qemu

Usage: $0 [options] usb
NOT IMPLEMENTED YET. Creates a ${MAGIC_NAME} usb tar.gz.

options:
        -h, --help      print this message and exit

EOF
        exit 1
}


clean_target()
{
# Prepare target folder
MAGIC_TARGET="./test_fs"


if [ -d  $MAGIC_TARGET ] ; then
  # If target folder exists we should delete it and create it again.
  rm -Rf $MAGIC_TARGET
  mkdir -p $MAGIC_TARGET
else
  # If target folder does exit we create it.
  mkdir -p $MAGIC_TARGET;
fi
}

common_setup()
{

# Copy /boot contents to to FS_TARGET
mkdir -p ${FS_TARGET}/boot
if cp -R ./boot/* ${FS_TARGET}/boot ; then
echo -e "GREAT!: We could copy ../boot folder into ${FS_TARGET} folder\n";
else
echo -e "ERROR: We could NOT copy ../boot folder into ${FS_TARGET} folder\n";
exit 1;
fi

# Update root menu.lst
#
#cp ./root_cd_menu.lst ../boot/grub/menu.lst

}

##############################################################
#########           MAIN PROGRAM	######################
##############################################################

case "$1" in
        -h | --help)
                usage
                ;;
	cdromiso)
		if [ ! -n "$2" ] ; then
		echo -e "ERROR:\n";
		echo -e "Please specify the ISO where you want to store ${MAGIC_DISTRONAME}\n"
		echo -e "e.g. \n"
		echo -e "$0 cdromiso /tmp/sgdplusknx_0.xx.iso\n"
		exit 1
		else
		clean_target
		common_setup
		./make_sgd.sh fs
		./make_slb.sh fs
		cp -R root/* $MAGIC_TARGET
		cp ./sgdplus.lst ${MAGIC_TARGET}/boot/grub/menu.lst
			if mkisofs -R -b boot/grub/stage2_eltorito -no-emul-boot -boot-load-size 4 -boot-info-table -o "$2" "$MAGIC_TARGET" ;
			then
			echo -e "GREAT!: We could make the iso file with mkisofs from $MAGIC_TARGET folder\n";
			else
			echo -e "ERROR: We could make the iso file with mkisofs from $MAGIC_TARGET folder. Check permissions.\n";
			exit 1;
			fi		
		fi
		;;


	fakecdromiso)
		if [ ! -n "$2" ] ; then
		echo -e "ERROR:\n";
		echo -e "Please specify the ISO where you want to store ${MAGIC_DISTRONAME}\n"
		echo -e "e.g. \n"
		echo -e "$0 cdromiso /tmp/sgdplusknx_0.xx.iso\n"
		exit 1
		else
		clean_target
		common_setup
		./make_sgd.sh fs
		./make_slb.sh fs
		cp ./sgdplus.lst ${MAGIC_TARGET}/boot/grub/menu.lst
			if mkisofs -R -b boot/grub/stage2_eltorito -no-emul-boot -boot-load-size 4 -boot-info-table -o "$2" "$MAGIC_TARGET" ;
			then
			echo -e "GREAT!: We could make the iso file with mkisofs from $MAGIC_TARGET folder\n";
			else
			echo -e "ERROR: We could make the iso file with mkisofs from $MAGIC_TARGET folder. Check permissions.\n";
			exit 1;
			fi		
		fi
		;;
		
	qemutest)
		SDG_TEST_ISO_FILE="./sgd_test.iso"
		$0 cdromiso ${SDG_TEST_ISO_FILE}
		qemu -boot d -cdrom ${SDG_TEST_ISO_FILE}
		rm ${SDG_TEST_ISO_FILE}
		;;
	fakeqemutest)
		SDG_TEST_ISO_FILE="./sgd_test.iso"
		$0 fakecdromiso ${SDG_TEST_ISO_FILE}
		qemu -boot d -cdrom ${SDG_TEST_ISO_FILE}
		rm ${SDG_TEST_ISO_FILE}
		;;

	usb)
		echo -e "USB BUILD:\n"
		echo -e "This feature is not finished yet.\n"
		echo -e "Complain to the SGD mailing list.\n"
		exit 1
		;;



        *)
		usage
                ;;
esac
