#!/bin/bash
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

set -x
set -v


usage()
{
        cat <<EOF

No options: Builds source code tar.bz2 in ../

options:
        -h, --help      print this message and exit

EOF
        exit 1
}


build_the_source_code () {

##############################################################
#########           MAIN PROGRAM	######################
##############################################################

# Get Super Grub Disk version

MAGIC_RELEASE="release"
MAGIC_FOLDER="dev_sgd"
MAKE_SOURCE_CODE_FOLDER=`pwd`

MAKE_SOURCE_CODE_FOLDER_FILENAME=${MAKE_SOURCE_CODE_FOLDER##*/};
MAKE_SOURCE_CODE_FOLDER_PATH=${MAKE_SOURCE_CODE_FOLDER%/*};
PUBLIC_MAKE_SOURCE_CODE_FOLDER_FILENAME="public_""${MAKE_SOURCE_CODE_FOLDER_FILENAME}"
PUBLIC_MAKE_SOURCE_CODE_FOLDER="${MAKE_SOURCE_CODE_FOLDER_PATH}""/""${PUBLIC_MAKE_SOURCE_CODE_FOLDER_FILENAME}"

SGD_SOURCE_CODE_STRING="source_code"
SGD_STRING="super_grub_disk"

# Check magic folder

	if [ ! -d "$MAGIC_FOLDER" ] ;   then
	echo -e "ERROR: $MAGIC_FOLDER folder is inaccesible\n";
	echo -e "ABORTING!";
	exit 1;
	fi

# Get data from configure
cd "${MAGIC_FOLDER}"
source ./sgrub_configure.sh
read_sgd_version

PUBLIC_TAR_BZ2_FILENAME="public""_""$SGD_STRING""_""$SGD_SOURCE_CODE_STRING""_""${SGD_VERSION}"".tar.bz2"

# Thanks to ./sgrub_configure.sh SGD_VERSION is set with the SGD version
# Go back to make source code folder
cd "$MAKE_SOURCE_CODE_FOLDER"
# Let's go outside the source code path
cd ..
OUTSIDE_SOURCE_CODE_PATH=`pwd`
# Let's clone our production system
cp -a "$MAKE_SOURCE_CODE_FOLDER" "${PUBLIC_MAKE_SOURCE_CODE_FOLDER}"
# Let's purge the public release
cd "${PUBLIC_MAKE_SOURCE_CODE_FOLDER}"
./clean.sh
# Go back to the outside source code so that we can do the public tar.bz2 file.
cd "$OUTSIDE_SOURCE_CODE_PATH"


tar -cjf "${PUBLIC_TAR_BZ2_FILENAME}" "${PUBLIC_MAKE_SOURCE_CODE_FOLDER_FILENAME}""/"
# Let's remove the unneed public temporal folder
rm -Rf "${PUBLIC_MAKE_SOURCE_CODE_FOLDER}"

# Let's move our file to the release folder on its correct place.
MAGIC_FTP_PATH="$MAKE_SOURCE_CODE_FOLDER""/""$MAGIC_RELEASE""/""download/source_code/sgd/"

if [ ! -d "$MAGIC_FTP_PATH" ] ; then
mkdir -p "$MAGIC_FTP_PATH"
fi


mv "${PUBLIC_TAR_BZ2_FILENAME}" "$MAGIC_FTP_PATH"

# Go back to current folder

cd "$MAKE_SOURCE_CODE_FOLDER"

# Nothing to do

}


case "$1" in
        -h | --help)
                usage
                ;;


        *)
		build_the_source_code
                ;;
esac